from pydantic import BaseModel


class CreateRequest(BaseModel):
    doc: str
    index: int
    collection: str


class ReadRequest(BaseModel):
    q: str
    docs: list[str] | None = None
    collection: str | None = None
    k: int


class UpdateRequest(BaseModel):
    index: int
    collection: str
    doc: str | None = None


class DeleteRequest(BaseModel):
    index: int
    collection: str



from sentence_transformers import SentenceTransformer
from torch import cuda
from torch.backends import mps


class App:
    _instance: "App" = None

    def __new__(cls) -> "App":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False

        return cls._instance

    def __init__(self) -> None:
        if not self._initialized:
            self._initialized = True

            self.model = None
            self.allow_origins = []
            self._default_encode = self.encode.__get__(
                self, App
            )  # Store reference to default implementation

    def run(self, host: str, port: int, allow_origins: list[str]) -> None:
        self.allow_origins = allow_origins

        # Setting a default model and encode function
        if not self.model:
            self.set_model(
                **{
                    "model_name_or_path": "avsolatorio/GIST-Embedding-v0",
                    "revision": None,
                }
            )
            self.set_model_device(get_device())

        # If the user did not override, set to a default
        if self.is_default_encode():

            def default_encode(
                text: str, convert_to_tensor: bool
            ) -> list[float]:
                return self.model.encode(
                    text, convert_to_tensor=convert_to_tensor
                )

            self.encode = default_encode

        import uvicorn

        from .router import api

        uvicorn.run(api, host=host, port=port)

    def set_model(self, **model_kwargs: dict) -> None:
        self.model = SentenceTransformer(**model_kwargs)
        self.model.eval()
        self.model.zero_grad()
        self.model.requires_grad_(False)

    def encode(
        self, text: str | list[str], *args: list, **kwargs: dict
    ) -> list[float] | list[list[float]]:
        """Default encode implementation that raises NotImplementedError"""

        raise NotImplementedError

    def is_default_encode(self) -> bool:
        """Check if current encode method is the default implementation"""

        return self.encode.__get__(self, App) == self._default_encode

    def set_model_device(self, device: str) -> None:
        self.model.to(device)


def get_device() -> str:
    if cuda.is_available():
        return "cuda"
    elif mps.is_available():
        return "mps"

    return "cpu"



from typing import Annotated

from fastapi import Depends
from sqlalchemy import JSON, Column
from sqlmodel import (
    Field,
    Index,
    Session,
    SQLModel,
    UniqueConstraint,
    create_engine,
    select,
)


class Embedding(SQLModel, table=True):
    id: int = Field(primary_key=True)
    index: int
    doc: str
    embedding: list[float] = Field(sa_column=Column(JSON))
    collection: str = Field(index=True)  # Add indexed collection field

    # Create composite primary key for id and collection
    __table_args__ = (Index("idx_collection_id", "collection", "id"),)


sqlite_file_name = "database.db"
sqlite_url = f"sqlite:///{sqlite_file_name}"

connect_args = {"check_same_thread": False}
engine = create_engine(sqlite_url, connect_args=connect_args)


def create_db_and_tables():
    SQLModel.metadata.create_all(engine)


def get_session():
    with Session(engine) as session:
        yield session


SessionDep = Annotated[Session, Depends(get_session)]


def create_embedding(
    session: Session,
    doc: str,
    embedding: list[float],
    index: int,
    collection: str,
):
    existing = session.exec(
        select(Embedding).where(
            (Embedding.index == index) & (Embedding.collection == collection)
        )
    ).first()

    if existing:
        raise ValueError(
            f"Embedding with index {index} in collection {collection} already exists"
        )

    db_embedding = Embedding(
        doc=doc, embedding=embedding, index=index, collection=collection
    )
    session.add(db_embedding)
    session.commit()
    session.refresh(db_embedding)
    return db_embedding


def read_collection(session: Session, collection: str) -> list[Embedding]:
    result = session.exec(
        select(Embedding)
        .where(Embedding.collection == collection)
        .order_by(Embedding.index)
    ).all()

    if not result:
        raise ValueError(f"Collection {collection} not found")

    return result


def update_embedding(
    session: Session,
    index: int,
    collection: str,
    doc: str,
    embedding_values: list[float],
) -> Embedding:
    statement = select(Embedding).where(
        (Embedding.index == index) & (Embedding.collection == collection)
    )
    db_embedding = session.exec(statement).first()
    if not db_embedding:
        raise ValueError(
            f"Embedding with index {index} in collection {collection} not found"
        )

    db_embedding.doc = doc
    db_embedding.embedding = embedding_values

    session.add(db_embedding)
    session.commit()
    session.refresh(db_embedding)
    return db_embedding


def delete_embedding(
    session: Session,
    index: int,
    collection: str,
) -> Embedding:
    statement = select(Embedding).where(
        (Embedding.index == index) & (Embedding.collection == collection)
    )
    db_embedding = session.exec(statement).first()
    if not db_embedding:
        raise ValueError(
            f"Embedding with index {index} in collection {collection} not found"
        )

    session.delete(db_embedding)
    session.commit()
    return db_embedding



from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from numpy import ndarray
from torch import Tensor, topk
from torch.nn.functional import cosine_similarity

from .db import (
    Embedding,
    SessionDep,
    create_db_and_tables,
    create_embedding,
    delete_embedding,
    read_collection,
    update_embedding,
)
from .main import App
from .schemas import CreateRequest, DeleteRequest, ReadRequest, UpdateRequest

api = FastAPI()
app = App()

api.add_middleware(
    CORSMiddleware,
    allow_origins=app.allow_origins,
    allow_methods=["POST", "GET", "PUT", "DELETE"],
    allow_headers=["Content-Type", "Authorization"],
    allow_credentials=True,
    max_age=3600,
)


@api.on_event("startup")
def on_startup():
    create_db_and_tables()


@api.post("/create")
def create_api(request: CreateRequest, session: SessionDep) -> dict:
    # try:
    created = create_embedding(
        session=session,
        doc=request.doc,
        embedding=tolist(app.encode(request.doc, convert_to_tensor=True)),
        index=request.index,
        collection=request.collection,
    )

    return {
        "status": "success",
    }


# except Exception as e:
#     raise HTTPException(status_code=500, detail=str(e))


@api.post("/read")
def read_api(request: ReadRequest, session: SessionDep) -> dict:
    try:
        if request.docs and request.collection:
            raise HTTPException(
                status_code=400,
                detail="Only one of 'docs' or 'collection' should be provided",
            )

        search = []
        if request.docs:
            search = enumerate(request.docs)
        if request.collection:
            search = (
                (db_embedding.index, db_embedding.doc)
                for db_embedding in read_collection(
                    session, request.collection
                )
            )

        search = tuple(search)

        if not search:
            raise HTTPException(
                status_code=400, detail="No documents to search"
            )

        q = app.encode(request.q, convert_to_tensor=True)
        docs = app.encode(
            tuple(item[1] for item in search), convert_to_tensor=True
        )
        similarities = cosine_similarity(q.unsqueeze(0), docs, dim=-1)
        scores, indices = topk(similarities, min(request.k, len(docs)))

        return {
            "scores": tolist(scores),
            "indices": [search[i][0] for i in tolist(indices)],
        }

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@api.put("/update")
def update_api(
    update_data: UpdateRequest,
    session: SessionDep,
) -> dict:
    try:
        updated = update_embedding(
            session,
            index=update_data.index,
            collection=update_data.collection,
            doc=update_data.doc,
            embedding_values=tolist(
                app.encode(update_data.doc, convert_to_tensor=True)
            ),
        )

        return {
            "status": "success",
        }

    except ValueError as e:
        raise HTTPException(status_code=404, detail=str(e))

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@api.delete("/delete")
def delete_api(
    delete_data: DeleteRequest,
    session: SessionDep,
) -> Embedding:
    try:
        deleted = delete_embedding(
            session,
            index=delete_data.index,
            collection=delete_data.collection,
        )

        return deleted

    except ValueError as e:
        raise HTTPException(status_code=404, detail=str(e))

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


def tolist(data: Tensor | ndarray) -> list:
    if isinstance(data, Tensor):
        return data.cpu().numpy().tolist()

    elif isinstance(data, ndarray):
        return data.tolist()

    raise ValueError("Data should be a Tensor or a numpy array.")



from setuptools import find_packages, setup

# easy_embed/easy_embed
with open("easy_embed/README.md", "r") as f:
    long_description = f.read()

setup(
    name="easy-embed-rafaelolal",  # easy_embed/src
    version="1.0.0",
    description="Simple self-hosted semantic search API",
    package_dir={"": "easy_embed"},  # easy_embed/easy_embed
    packages=find_packages(where="easy_embed"),  # easy_embed/easy_embed
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rafaelolal/csci-3485-final",
    author="Rafael Almeida",
    author_email="contact@ralmeida.dev",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "annotated-types==0.7.0",
        "anyio==4.7.0",
        "certifi==2024.12.14",
        "charset-normalizer==3.4.0",
        "click==8.1.7",
        "dnspython==2.7.0",
        "email_validator==2.2.0",
        "fastapi==0.115.6",
        "fastapi-cli==0.0.7",
        "filelock==3.16.1",
        "fsspec==2024.10.0",
        "h11==0.14.0",
        "httpcore==1.0.7",
        "httptools==0.6.4",
        "httpx==0.28.1",
        "huggingface-hub==0.26.5",
        "idna==3.10",
        "Jinja2==3.1.4",
        "joblib==1.4.2",
        "markdown-it-py==3.0.0",
        "MarkupSafe==3.0.2",
        "mdurl==0.1.2",
        "mpmath==1.3.0",
        "networkx==3.4.2",
        "numpy==2.2.0",
        "packaging==24.2",
        "pillow==11.0.0",
        "pydantic==2.10.3",
        "pydantic_core==2.27.1",
        "Pygments==2.18.0",
        "python-dotenv==1.0.1",
        "python-multipart==0.0.19",
        "PyYAML==6.0.2",
        "regex==2024.11.6",
        "requests==2.32.3",
        "rich==13.9.4",
        "rich-toolkit==0.12.0",
        "safetensors==0.4.5",
        "scikit-learn==1.6.0",
        "scipy==1.14.1",
        "sentence-transformers==3.3.1",
        "setuptools==75.6.0",
        "shellingham==1.5.4",
        "sniffio==1.3.1",
        "SQLAlchemy==2.0.36",
        "sqlmodel==0.0.22",
        "starlette==0.41.3",
        "sympy==1.13.1",
        "threadpoolctl==3.5.0",
        "tokenizers==0.21.0",
        "torch==2.5.1",
        "tqdm==4.67.1",
        "transformers==4.47.0",
        "typer==0.15.1",
        "typing_extensions==4.12.2",
        "urllib3==2.2.3",
        "uvicorn==0.34.0",
        "uvloop==0.21.0",
        "watchfiles==1.0.3",
        "websockets==14.1",
    ],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires="==3.12.8",
)
